package edu.neu.ccs.demeter.aplib;

import java.util.*;

/**
 * Like <tt>HashMap</tt> but the iterator is guaranteed to visit the
 * elements in the same order they were added.
 *
 * @see HashMap
 * @see OrderPreservingHashSet
 */
public class OrderPreservingHashMap extends HashMap {
  private transient ArrayList array;

  /**
   * Constructs a new, empty map with the specified initial 
   * capacity and the specified load factor. 
   *
   * @param      initialCapacity   the initial capacity of the HashMap.
   * @param      loadFactor        the load factor of the HashMap
   * @throws     IllegalArgumentException  if the initial capacity is less
   *               than zero, or if the load factor is nonpositive.
   */
  public OrderPreservingHashMap(int initialCapacity, float loadFactor) {
    super(initialCapacity, loadFactor);
    array = new ArrayList(initialCapacity);
  }

  /**
   * Constructs a new, empty map with the specified initial capacity
   * and default load factor, which is <tt>0.75</tt>.
   *
   * @param   initialCapacity   the initial capacity of the HashMap.
   * @throws    IllegalArgumentException if the initial capacity is less
   *              than zero.
   */
  public OrderPreservingHashMap(int initialCapacity) {
    super(initialCapacity);
    array = new ArrayList(initialCapacity);
  }

  /**
   * Constructs a new, empty map with a default capacity and load
   * factor, which is <tt>0.75</tt>.
   */
  public OrderPreservingHashMap() {
    super();
    array = new ArrayList(101);
  }

  /**
   * Constructs a new map with the same mappings as the given map.  The
   * map is created with a capacity of twice the number of mappings in
   * the given map or 11 (whichever is greater), and a default load factor,
   * which is <tt>0.75</tt>.
   */
  public OrderPreservingHashMap(Map t) {
    // We can't just call super(t) because that would start
    // inserting things before we get to construct the array.
    super(Math.max(2*t.size(), 11));
    array = new ArrayList(Math.max(2*t.size(), 11));
    putAll(t);
  }

  /**
   * Associates the specified value with the specified key in this map.
   * If the map previously contained a mapping for this key, the old
   * value is replaced.
   *
   * @param key key with which the specified value is to be associated.
   * @param value value to be associated with the specified key.
   * @return previous value associated with specified key, or <tt>null</tt>
   *	       if there was no mapping for key.  A <tt>null</tt> return can
   *	       also indicate that the HashMap previously associated
   *	       <tt>null</tt> with the specified key.
   */
  public Object put(Object key, Object value) {
    Object old = super.put(key, value);
    if (old != null)
      // FIXME: this is slow!
      array.remove(old);
    array.add(value);
    return old;
  }

  /**
   * Removes the mapping for this key from this map if present.
   *
   * @param key key whose mapping is to be removed from the map.
   * @return previous value associated with specified key, or <tt>null</tt>
   *	       if there was no mapping for key.  A <tt>null</tt> return can
   *	       also indicate that the map previously associated <tt>null</tt>
   *	       with the specified key.
   */
  public Object remove(Object key) {
    Object old = super.remove(key);
    if (old != null)
      // FIXME: this is slow!
      array.remove(old);
    return old;
  }

  /**
   * Removes all mappings from this map.
   */
  public void clear() {
    super.clear();
    array.clear();
  }

  /**
   * Returns a shallow copy of this <tt>OrderPreservingHashMap</tt>
   * instance: the keys and values themselves are not cloned.
   *
   * @return a shallow copy of this map.
   */
  public Object clone() {
    OrderPreservingHashMap clone = (OrderPreservingHashMap) super.clone();
    clone.array = (ArrayList) array.clone();
    return clone;
  }

  private transient Set entrySet = null;
  private transient Collection values = null;

  /**
   * Returns a collection view of the values contained in this map.  The
   * collection is backed by the map, so changes to the map are reflected in
   * the collection, and vice-versa.  The collection supports element
   * removal, which removes the corresponding mapping from this map, via the
   * <tt>Iterator.remove</tt>, <tt>Collection.remove</tt>,
   * <tt>removeAll</tt>, <tt>retainAll</tt>, and <tt>clear</tt> operations.
   * It does not support the <tt>add</tt> or <tt>addAll</tt> operations.
   * The collection iterator is guaranteed to visit the values in the same
   * order that they were added.
   *
   * @return a collection view of the values contained in this map.
   */
  public Collection values() {
    if (values==null) {
      values = new AbstractCollection() {
	  public Iterator iterator() {
	    return array.iterator();
	  }
	  public int size() {
	    return OrderPreservingHashMap.this.size();
	  }
	  public boolean contains(Object o) {
	    return containsValue(o);
	  }
	  public void clear() {
	    OrderPreservingHashMap.this.clear();
	  }
	};
    }
    return values;
  }

  /**
   * Returns a collection view of the mappings contained in this map.  Each
   * element in the returned collection is a <tt>Map.Entry</tt>.  The
   * collection is backed by the map, so changes to the map are reflected in
   * the collection, and vice-versa.  The collection supports element
   * removal, which removes the corresponding mapping from the map, via the
   * <tt>Iterator.remove</tt>, <tt>Collection.remove</tt>,
   * <tt>removeAll</tt>, <tt>retainAll</tt>, and <tt>clear</tt> operations.
   * It does not support the <tt>add</tt> or <tt>addAll</tt> operations.
   *
   * @return a collection view of the mappings contained in this map.
   * @see Map.Entry
   */
  public Set entrySet() {
    // FIXME: preserve order
    return super.entrySet();
  }

  /**
   * Returns an unmodifiable List view of the values in the map, in the
   * same order as they were added.
   *
   * @return an unmodifiable List view of the values.
   */
  public List valuesAsList() {
    return Collections.unmodifiableList(array);
  }
}
