package edu.neu.ccs.demeter.aplib;

import java.util.*;

/**
 * Like <tt>HashSet</tt> but the iterator is guaranteed to visit the
 * elements in the same order they were added.
 *
 * @see HashSet
 * @see OrderPreservingHashMap
 */

public class OrderPreservingHashSet extends HashSet {
  private transient ArrayList array;

  /**
   * Constructs a new, empty set; the backing <tt>HashMap</tt> instance has
   * default capacity and load factor, which is <tt>0.75</tt>.
   */
  public OrderPreservingHashSet() {
    super();
    array = new ArrayList();
  }

  /**
   * Constructs a new set containing the elements in the specified
   * collection.  The capacity of the backing <tt>HashMap</tt> instance is
   * twice the size of the specified collection or eleven (whichever is
   * greater), and the default load factor (which is <tt>0.75</tt>) is used.
   */
  public OrderPreservingHashSet(Collection c) {
    // The superclass constructor calls add() on each object, so we
    // need to make the array first, but the super() call has to be
    // the syntactically first thing in the body.  So we have to do
    // the adding ourselves.
    super(Math.max(2*c.size(), 11));
    // This is lame-- there's no way to cache this computation, so we
    // have to do it again!
    array = new ArrayList(Math.max(2*c.size(), 11));
    addAll(c);
  }

  /**
   * Constructs a new, empty set; the backing <tt>HashMap</tt> instance has
   * the specified initial capacity and the specified load factor.
   *
   * @param      initialCapacity   the initial capacity of the hash map.
   * @param      loadFactor        the load factor of the hash map.
   * @throws     IllegalArgumentException if the initial capacity is less
   *             than zero, or if the load factor is nonpositive.
   */
  public OrderPreservingHashSet(int initialCapacity, float loadFactor) {
    super(initialCapacity, loadFactor);
    array = new ArrayList(initialCapacity);
  }

  /**
   * Constructs a new, empty set; the backing <tt>HashMap</tt> instance has
   * the specified initial capacity and default load factor, which is
   * <tt>0.75</tt>.
   *
   * @param      initialCapacity   the initial capacity of the hash table.
   * @throws     IllegalArgumentException if the initial capacity is less
   *             than zero.
   */
  public OrderPreservingHashSet(int initialCapacity) {
    super(initialCapacity);
    array = new ArrayList(initialCapacity);
  }
    
  /**
   * Returns an iterator over the elements in this set.  The elements
   * are returned in the same order they were added.
   *
   * @return an Iterator over the elements in this set.
   */
  public Iterator iterator() { return array.iterator(); }

  /**
   * Adds the specified element to this set if it is not already
   * present.
   *
   * @param o element to be added to this set.
   * @return <tt>true</tt> if the set did not already contain the specified
   * element.
   */
  public boolean add(Object o) {
    boolean added = super.add(o);
    if (added) array.add(o);
    return added;
  }

  /**
   * Removes the given element from this set if it is present.
   *
   * @param o object to be removed from this set, if present.
   * @return <tt>true</tt> if the set contained the specified element.
   */
  public boolean remove(Object o) {
    boolean removed = super.remove(o);
    // FIXME: this is slow!
    if (removed) array.remove(o);
    return removed;
  }

  /**
   * Removes all of the elements from this set.
   */
  public void clear() {
    super.clear();
    array.clear();
  }

  /**
   * Returns a shallow copy of this <tt>OrderPreservingHashSet</tt>
   * instance: the elements themselves are not cloned.
   *
   * @return a shallow copy of this set.
   */
  public Object clone() {
    OrderPreservingHashSet newSet = (OrderPreservingHashSet) super.clone();
    newSet.array = (ArrayList) array.clone();
    return newSet;
  }

  /**
   * Returns an unmodifiable List view of the elements in the set, in the
   * same order as they were added.
   *
   * @return an unmodifiable List view.
   */
  public List asList() {
    return Collections.unmodifiableList(array);
  }
}
